/* -*-C-*- tscanf.c */
/*-->tscanf*/
/**********************************************************************/
/****************************** tscanf ********************************/
/**********************************************************************/

/***********************************************************************
tscanf() provides sscanf-like support for simplifying the code needed to
parse the various forms of foreign \special commands supported by the
DVI driver family.  The code has been adapted from sscanf.c, which is
also part of the DVI driver family code.

Unfortunately, sscanf() is missing some features that are useful in TeX
software:

	(a) "key = value" and "key=value" cannot both be recognized by
	just one of the formats "%s = %s" and "%s=%s", making it hard to
	support optional whitespace in assignments;

	(b) quoted strings containing doubled quotes (representing a
	single one) or whitespace cannot be recognized;

	(c) there is no way to handle TeX dimensions;

	(d) there is no way to handle the TeXtures convention that an
	integer nnnn represents the floating-point value nnnn/1000,
	while nnnn. or nnnn.0 is an unscaled floating-point value.

tscanf() implements all sscanf() format items exactly, except for %p
(pointer) and %L (long double), which are not supported.  tscanf()
also provides these additional format items:

	%S	like %s, except that = is excluded from the match

	%q	quoted string: the surrounding quotes are not included
		in the returned string, and inner doubled quotes are
		reduced to single quotes

	%T	TeX dimension (converted to floating-point TeX points)

	%t	TeXtures floating-point number

The standard *, nnn, h, and l modifiers are also available for these
items.

tscanf() requires support for a variable number of arguments, and the
comments in the companion sscanf.c file describe the difficulties that
this poses for portable software.  Fortunately, implementations of
Standard C are now widely available (in 1995), so this is less of a
problem than it used to be: only two of the dozen or so flavors of
UNIX systems that the author routinely works on still lack a
vendor-provided Standard C compiler, and GNU gcc is available on one
of them.

A (too) simple test program is contained in this file, and will be
compiled if the symbol TEST is defined at compile time.  When run, it
should produce no output, unless there are errors detected.

The implementation of the format string processing includes:

	(1) whitespace, which matches zero or more whitespace
	characters.

	(2) %<format-item>
		where <format-item> =
		*               optional assignment suppression
		nnn             optional digit string defining maximum
				field width (may not exceed MAX_FIELD_WIDTH
				below)
		h               optional short modifier for d, i, or n items
		l               optional long modifier for d, i, or n items
		[...]           character set to read
		%               literal percent
		c               character input
		d               decimal integer
		e,f,g,E,G       floating-point conversion
		i               C-style integer input
		n               input character count
		q		quoted string
		s               character string
		S		like %s, but excludes = from match
		t		TeXtures floating-point value: if an
				integer is found, its value is divided
				by 1000; otherwise, conversion is like
				e,f,g,E,G format items.
		T		TeX floating-point dimension (in TeX pt)
		u               unsigned decimal integer

	(3) other               other literal character

The C Standard specifies that the interpretation of a - in a character
set is undefined; this implementation follows the practice adopted by
Sun Microsystems of interpreting a - as a range indicator, so that
[a-eX-Z-] is equivalent to [abcdeXYZ-] in a character set where the
letters are contiguous (this is not the case in the EBCDIC character
set).

Not yet implemented (and not required in DVI driver applications) are:

	<format-item> =
		p               pointer input (system-dependent: needs to
				be coordinated with printf() family output)
	<modifier-item> =
		L		corresponding argument is of type
				`long double', which few C implementations
				support.

[26-Oct-1996]	-- add support for multiple terminal %[lh]n format items,
		   and add test 37 in test program to check for correct
		   handling of such a format
[20-Dec-1995]	-- add forgotten support for h and l modifiers of %n, and
		   extra tests 1.3 .. 1.6 in main()
[16-Jul-1995]	-- add test for NULL fmt and str
[30-Jun-1995]	-- adapted sscanf.c code to make tscanf.c
[02-Jul-1989]	-- original version of sscanf.c
***********************************************************************/

#include "headers.h"

#if defined(va_dcl) && defined(HAVE_VARARGS_H) && !defined(STDC)
/* va_dcl is used in old varargs code, but some systems (e.g. DEC Alpha OSF/1)
define it in stdarg.h too, sigh... */
#define HAVE_VARARGS_STYLE
#endif


RCSID("$Id: tscanf.c,v 1.1 1997/09/01 22:26:35 beebe Exp beebe $")

#define Tolower(c)	(isupper(c) ? tolower(c) : (c))
#define isdimen(s,d)	((Tolower((s)[0]) == (d)[0]) && \
			 (Tolower((s)[1]) == (d)[1]))

#define MAXCHARS	256		/* size of host character set */

#define MAX_FIELD_WIDTH	32767		/* maximum field width supported: */
					/* should not exceed the smallest */
					/* maximum integer represented by */
					/* type size_t on any architecture */

static const char	*flush_space ARGS((const char* s_));

static const char*
#if STDC
flush_space(
const char			*s
)
#else /* NOT STDC */
flush_space(s)
const char			*s;
#endif /* STDC */
{
    while (*s && isspace(*s))
	++s;
    return (s);
}

#ifdef __GNUC__
#define GNUC 1
#else /* __GNUC__ */
#define GNUC 0
#endif /* __GNUC__ */

#if    (STDC || GNUC)
/* Some compilers (e.g. Turbo C on PC DOS) will not accept the ellipsis
except in an Standard C-style header. */
int
tscanf(const char *str,const char *fmt,...)
#else /* NOT (STDC || GNUC) */
int
tscanf(str,fmt,va_alist)
const char			*str;
const char			*fmt;
#if defined(HAVE_VARARGS_STYLE)
va_dcl
#endif /* defined(HAVE_VARARGS_STYLE) */
#endif /* (STDC || GNUC) */
{
    int				inset;	/* char is in character set [...] */
    va_list			ap;	/* argument list pointer */
    int				assignment_suppress;
    int				base;	/* number base for u, o, and x items */
    int				charset[MAXCHARS];
					/* character set in [...] item */

#define INSET(c)	charset[(unsigned int)(0xff & c)]

    char			*endptr;/* pointer to end of strxxx() parse */
    int				errors;	/* error count */
    int				item_type; /* type modifier TYPE_xxx code */

#define TYPE_NORMAL	0		/* type codes for item_type */
#define	TYPE_SHORT	1
#define TYPE_LONG	2

    int				k;	/* loop index */
    int				kset;	/* number of scan set members */
    size_t			max_field_width;
    int				more;	/* non-zero if more of format item */
    int				n;	/* number of successful conversions */
    int				nread;	/* number of source characters read */
    const char			*org_str; /* copy of original str */
    char			*p;	/* temporary buffer pointer */
    char			*q;	/* temporary buffer pointer */
    double			vtmp_double;/* local value result temporaries */
    long			vtmp_long;
    char			*vtmp_string = (char*)NULL;
    unsigned long		vtmp_ulong;

#if defined(HAVE_VARARGS_STYLE)
    va_start(ap);			/* set ap to point to first vararg */
#else /* !defined(HAVE_VARARGS_STYLE) */
    va_start(ap,fmt);			/* set ap to point to first vararg */
#endif /* defined(HAVE_VARARGS_STYLE) */

    if ((fmt == (const char*)NULL) || (str == (const char*)NULL))
	return (EOF);			/* error: cannot dereference NULL */

    assignment_suppress = 0;
    max_field_width = 0;
    item_type = TYPE_NORMAL;
    org_str = str;			/* need for 'n' item */
    p = (char*)NULL;
    for (errors = 0, n = 0; *fmt && *str && (errors == 0); fmt++,str++)
    {
	if (isspace(*fmt))
	{
	    str = flush_space(str);
	    str--;			/* because loop end advances str */
	    fmt = flush_space(fmt);
	    fmt--;			/* because loop end advances fmt */
	}
	else if (*fmt == '%')
	{
	    ++fmt;
	    for (more = 1; more && (errors == 0); )
	    {
		more = 0;
		switch (*fmt)
		{
		case '%':
		    if (*str != '%')
			errors++;
		    break;

		case '*':
		    if (assignment_suppress)
		    {
			errors++;	/* only one * allowed per item */
			break;
		    }
		    assignment_suppress = 1;
		    more = 1;
		    ++fmt;
		    break;

		case '[':
		    ++fmt;
		    if (*fmt == '^')		/* then we have a negated set */
		    {
			inset = 0;
			++fmt;
		    }
		    else
			inset = 1;
		    for (k = 0; k < MAXCHARS; ++k) /* clear character set */
			INSET(k) = !inset;
		    kset = 0;
		    if (*fmt == ']') /* special case: right-bracket in set */
		    {
			INSET(*fmt) = inset;
			++fmt;
			++kset;
		    }
		    while (*fmt && (*fmt != ']')) /* add set members */
		    {
			++kset;
			if ( (kset > 1)		/* not first set char */
			    && (*fmt == '-')	/* AND range selector */
			    && (*(fmt+1) != ']') /* AND not last set char */
			    && (*(fmt-1) <= *(fmt+1)) /* AND init <= final */
			    )
			{ /* support common range extension to Standard C */
			    for (k = (int)(*(fmt-1)); k <= (int)(*(fmt+1)); ++k)
				INSET(k) = inset;
			    ++fmt;		/* advance over minus char */
			    ++kset;
			}
			else
			{
			    INSET(*fmt) = inset;
			    ++fmt;
			}
		    }
		    if (*fmt != ']')
		    {
			errors++;
			break;
		    }
		    if (!assignment_suppress)
			vtmp_string = va_arg(ap,char*);
		    if (max_field_width == 0)
			max_field_width = MAX_FIELD_WIDTH;
		    for (nread = 0;
			 (*str && INSET(*str) && (max_field_width > 0));
			 ++nread, ++str, --max_field_width)
		    {
			if (!assignment_suppress)
			    *vtmp_string++ = *str;
		    }
		    if ((nread > 0) && !assignment_suppress)
		    {
			*vtmp_string = '\0';
			n++;
		    }
		    --str;
		    break;

		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		    if (max_field_width > 0)
		    {
			errors++;	/* only one field width per item */
			break;
		    }
		    more = 1;
		    for (max_field_width = 0; *fmt && isdigit(*fmt); ++fmt)
			max_field_width = 10*max_field_width +
			    (int)((*fmt) - '0');
		    if (max_field_width == 0)
		    {
			more = 0;
			errors++;	/* 0 field width is illegal */
		    }
		    break;

		case 'c':
		    if (max_field_width == 0)
			max_field_width = 1;
		    if (!assignment_suppress)
			vtmp_string = va_arg(ap,char*);
		    while (*str && (max_field_width > 0))
		    {
			if (!assignment_suppress)
			    *vtmp_string++ = *str;
			++str;
			--max_field_width;
		    }
		    if (!assignment_suppress) /* NB: no NUL terminator! */
			n++;
		    --str;
		    break;

		case 'd':
		case 'i':
		    base = (*fmt == 'i') ? 0 : 10;
		    errno = 0;
		    if (max_field_width > 0)
		    {
			p = (char*)malloc(max_field_width + 1);
			if (p == (char*)NULL)
			{
			    perror("tscanf():  Cannot allocate memory");
			    exit(EXIT_FAILURE);
			}
			(void)strcpy(p,str);
			p[max_field_width] = '\0';
			vtmp_long = (long)strtol(p,&endptr,base);
			FREE(p);
		    }
		    else
		    {
			p = (char*)str;
			vtmp_long = (long)strtol(p,&endptr,base);
		    }
		    if (errno != 0)
			errors++;
		    else
		    {
			str += (int)(endptr - p) - 1;
			switch (item_type)
			{
			case TYPE_NORMAL:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,int*) = (int)vtmp_long;
				n++;
			    }
			    break;
			case TYPE_LONG:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,long*) = (long)vtmp_long;
				n++;
			    }
			    break;
			case TYPE_SHORT:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,short*) = (short)vtmp_long;
				n++;
			    }
			    break;
			default:
			    errors++;
			}
			item_type = TYPE_NORMAL;
		    }
		    break;

		case 'e':		    /* fall through */
		case 'E':		    /* fall through */
		case 'f':		    /* fall through */
		case 'g':		    /* fall through */
		case 'G':		    /* fall through */
		case 't':		    /* fall through */
		case 'T':
		    errno = 0;
		    if (max_field_width > 0)
		    {
			p = (char*)malloc(max_field_width + 1);
			if (p == (char*)NULL)
			{
			    perror("tscanf():  Cannot allocate memory");
			    exit(EXIT_FAILURE);
			}
			(void)strcpy(p,str);
			p[max_field_width] = '\0';
			vtmp_double = (double)strtod(p,&endptr);
			FREE(p);
		    }
		    else
		    {
			p = (char*)str;
			vtmp_double = (double)strtod(p,&endptr);
		    }
		    if (errno != 0)
			errors++;
		    else
		    {
			str += (int)(endptr - p) - 1;
			switch (*fmt)
			{
			case 'T':	/* TeX dimension in pt*/
			    if (isdimen(str+1,"bp")) /* big points */
				vtmp_double *= 72.27 / 72.0;
			    else if (isdimen(str+1,"in")) /* inches */
				vtmp_double *= 72.27;
			    else if (isdimen(str+1,"cm")) /* centimeters */
				vtmp_double *= 72.27 / 2.54;
			    else if (isdimen(str+1,"pt")) /* points */
				/* NO-OP */;
			    else if (isdimen(str+1,"cc")) /* cicero */
				vtmp_double *= 12.0 * (1238.0 / 1157.0);
			    else if (isdimen(str+1,"dd")) /* didot points */
				vtmp_double *= (1238.0 / 1157.0);
			    else if (isdimen(str+1,"mm")) /* millimeters */
				vtmp_double *= 72.27 / 25.4;
			    else if (isdimen(str+1,"pc")) /* pica */
				vtmp_double *= 12.0;
			    else if (isdimen(str+1,"sp")) /* scaled points */
				vtmp_double /= 65536.0;
			    else	/* unknown units -- error */
				errors++;
			    if (errors == 0)
				str += 2; /* point to last char of units */
			    break;

			case 't':	/* TeXtures number: missing decimal */
					/* means divide by 1000 */
			    q = strchr(p,'.');
			    if (q == (const char*)NULL)	/* no . found */
				vtmp_double /= 1000.0;
			    else if ((q != (const char*)NULL) &&
				     (q >= endptr)) /* . found after field */
				vtmp_double /= 1000.0;
			    break;

			default:
			    /* NO-OP */;
			}
			if (errors > 0)	/* must have unknown %T units */
			    break;
			switch (item_type)
			{
			case TYPE_NORMAL:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,float*) = (float)vtmp_double;
				n++;
			    }
			    break;
			case TYPE_LONG:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,double*) = (double)vtmp_double;
				n++;
			    }
			    break;
			default:
			    errors++;
			}
			item_type = TYPE_NORMAL;
		    }
		    break;

		case 'h':
		    if (item_type != TYPE_NORMAL)
		    {
			errors++;	/* only one type modifier per item */
			break;
		    }
		    item_type = TYPE_SHORT;
		    more = 1;
		    ++fmt;
		    break;

		case 'l':
		    if (item_type != TYPE_NORMAL)
		    {
			errors++;	/* only one type modifier per item */
			break;
		    }
		    item_type = TYPE_LONG;
		    more = 1;
		    ++fmt;
		    break;

		case 'n': /* NB: NO increment of assignment count in n */
		    vtmp_long = (long)(str - org_str);
		    switch (item_type)
		    {
		    case TYPE_NORMAL:
			if (!assignment_suppress)
			    *va_arg(ap,int*) = (int)vtmp_long;
			break;
		    case TYPE_LONG:
			if (!assignment_suppress)
			    *va_arg(ap,long*) = vtmp_long;
			break;
		    case TYPE_SHORT:
			if (!assignment_suppress)
			    *va_arg(ap,short*) = (short)vtmp_long;
			break;
		    default:
			errors++;
		    }
		    item_type = TYPE_NORMAL;
		    --str;		/* counteract loop-ending ++str */
		    break;

		case 'q':
		    if (!assignment_suppress)
			vtmp_string = va_arg(ap,char*);
		    str = flush_space(str);
		    if (max_field_width == 0)
			max_field_width = MAX_FIELD_WIDTH;
		    if (*str != '"')
			break;
		    str++;		/* advance past opening quote */
		    while (*str && (max_field_width > 0))
		    {
			if (*str == '"')
			{
			    if (*(str+1) != '"')
				break;	/* end of string */
			    else	/* reduce embedded "" to " */
				++str;
			}
			if (!assignment_suppress)
			    *vtmp_string++ = *str;
			++str;
			--max_field_width;
		    }
		    if (!assignment_suppress)
		    {
			*vtmp_string = '\0';
			n++;
		    }
		    break;

		case 'S':		/* fall through */
		case 's':
		    if (!assignment_suppress)
			vtmp_string = va_arg(ap,char*);
		    str = flush_space(str);
		    if (max_field_width == 0)
			max_field_width = MAX_FIELD_WIDTH;
		    while (*str && (max_field_width > 0))
		    {
			if (isspace(*str))
			    break;
			if ((*fmt == 'S') && (*str == '='))
			    break;
			if (!assignment_suppress)
			    *vtmp_string++ = *str;
			++str;
			--max_field_width;
		    }
		    if (!assignment_suppress)
		    {
			*vtmp_string = '\0';
			n++;
		    }
		    --str;
		    break;

		case 'o':		    /* fall through */
		case 'u':		    /* fall through */
		case 'X':		    /* fall through */
		case 'x':
		    if (*fmt == 'o')
			base = 8;
		    else if (*fmt == 'u')
			base = 10;
		    else
			base = 16;
		    errno = 0;
		    if (max_field_width > 0)
		    {
			p = (char*)malloc(max_field_width + 1);
			if (p == (char*)NULL)
			{
			    perror("tscanf():  Cannot allocate memory");
			    exit(EXIT_FAILURE);
			}
			(void)strcpy(p,str);
			p[max_field_width] = '\0';
			vtmp_ulong = strtoul(p,&endptr,base);
			FREE(p);
		    }
		    else
		    {
			p = (char*)str;
			vtmp_ulong = strtoul(p,&endptr,base);
		    }
		    if (errno != 0)
			errors++;
		    else
		    {
			str += (int)(endptr - p) - 1;
			switch (item_type)
			{
			case TYPE_NORMAL:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,unsigned int*) =
				    (unsigned int)vtmp_ulong;
				n++;
			    }
			    break;
			case TYPE_LONG:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,unsigned long*) = vtmp_ulong;
				n++;
			    }
			    break;
			case TYPE_SHORT:
			    if (!assignment_suppress)
			    {
				*va_arg(ap,unsigned short*) =
				    (unsigned short)vtmp_ulong;
				n++;
			    }
			    break;
			default:
			    errors++;
			}
			item_type = TYPE_NORMAL;
		    }
		    break;

		default:
		    errors++;
		    break;
		}			/* end switch (*fmt) */
	    }				/* end for (more...) */
	    max_field_width = 0;
	    assignment_suppress = 0;
	    more = 0;
	}
	else				/* ordinary character: exact match */
	{
	    if (*fmt != *str)
	    {
		++errors;
		break;			/* error exit */
	    }
	}
    }					/* end for (n = ...) */

    /* Final processing: handle any trailing " %[hl]n" in the format,
       There may be multiple such format items, so we need a loop here. */
    fmt = flush_space(fmt);
    while (*fmt == '%')
    {
	switch (*(fmt + 1))
	{
	case 'h':
	    item_type = TYPE_SHORT;
	    fmt++;
	    break;
	case 'l':
	    item_type = TYPE_LONG;
	    fmt++;
	    break;
	default:
	    item_type = TYPE_NORMAL;
	    break;
	}

	if (*(fmt + 1) == 'n')
	{
	    vtmp_long = (long)(str - org_str);
	    switch (item_type)
	    {
	    case TYPE_NORMAL:
		if (!assignment_suppress)
		    *va_arg(ap,int*) = (int)vtmp_long;
		break;
	    case TYPE_LONG:
		if (!assignment_suppress)
		    *va_arg(ap,long*) = vtmp_long;
		break;
	    case TYPE_SHORT:
		if (!assignment_suppress)
		    *va_arg(ap,short*) = (short)vtmp_long;
		break;
	    default:
		errors++;
	    }
	    fmt += 2;			/* advance past 'n' */
	}
	else
	    break;
	fmt = flush_space(fmt);
    }
    va_end(ap);				/* done processing varargs */

    if (errors > 0)
	return (n == 0 ? EOF : n);
    else
	return (n);
}

#ifdef TEST

#define ERROR(fmt,u,v)	(\
	error_count++,\
	(void)fprintf(stderr,fmt,u,v),\
	fputc('\n',stderr),\
	(void)fflush(stderr)\
	)

#define MAXLINE 256
#define UNSET	(-9999)

int		main ARGS((int argc_, char* argv_[]));
static void	check_pages ARGS((int got_, int expect_, long pages_[10],
				  int next_char_));
static void	clear_pages ARGS((long pages_[10]));
static void	clear_string ARGS((char *s_, size_t n_));

static int	error_count = 0;

int
#if STDC
main(
int argc,
char* argv[]
)
#else /* NOT STDC */
main(argc,argv)
int argc;
char* argv[];
#endif /* STDC */
{
    float	f;
    float	fdimen[9];
    double	d;
    double	ddimen[9];
    double	dllx;
    double	dlly;
    double	durx;
    double	dury;
    long	llx;
    long	lly;
    long	urx;
    long	ury;
    int		idimen[9];
    int		llx_int;
    int		lly_int;
    int		urx_int;
    int		ury_int;
    short	llx_short;
    short	lly_short;
    short	urx_short;
    short	ury_short;
    int		i;
    int		k;
    int		n;
    long	l;
    long	n_long;
    long	pages[10];
    char	s[MAXLINE];
    char	s1[MAXLINE];
    char	s2[MAXLINE];
    char	s3[MAXLINE];
    char	s4[MAXLINE];
    const char *str;
    short	n_short;

    n = (int)UNSET;
    k = tscanf(str = "%%BoundingBox","%%%%BoundingBox%n",&n);
    if ((k != 0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 1: tscanf() -> %d:%s",k,"");

    llx = lly = urx = ury = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "%%BoundingBox: 123 456 789 987",
	"%%%%BoundingBox:%ld %ld %ld %ld%n",&llx,&lly,&urx,&ury,&n);
    if ((k != 4) || (llx != 123L) || (lly != 456L) || (urx != 789L) ||
	(ury != 987L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 1.1: tscanf() -> %d:%s",k,"");

    dllx = dlly = durx = dury = (double)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "%%BoundingBox: 123 456 789 987",
	"%%%%BoundingBox:%lf %lf %lf %lf%n",&dllx,&dlly,&durx,&dury,&n);
    if ((k != 4) || (dllx != 123.0) || (dlly != 456.0) || (durx != 789.0) ||
	(dury != 987.0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 1.2: tscanf() -> %d:%s",k,"");

    n_short = (short)UNSET;
    k = tscanf(str = "%%BoundingBox","%%%%BoundingBox%hn",&n_short);
    if ((k != 0) || (n_short == (short)UNSET) || (str[n_short] != '\0'))
	ERROR("Test 1.3: tscanf() -> %d:%s",k,"");

    n_long = (long)UNSET;
    k = tscanf(str = "%%BoundingBox","%%%%BoundingBox%ln",&n_long);
    if ((k != 0) || (n_long == (long)UNSET) || (str[n_long] != '\0'))
	ERROR("Test 1.4: tscanf() -> %d:%s",k,"");

    llx_int = lly_int = urx_int = ury_int = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "%%BoundingBox: 123 456 789 987",
	"%%%%BoundingBox:%d %d %d %d%n",&llx_int,&lly_int,&urx_int,&ury_int,&n);
    if ((k != 4) || (llx_int != 123) || (lly_int != 456) || (urx_int != 789) ||
	(ury_int != 987) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 1.5: tscanf() -> %d:%s",k,"");

    llx_short = lly_short = urx_short = ury_short = (short)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "%%BoundingBox: 123 456 789 987",
	"%%%%BoundingBox:%hd %hd %hd %hd%n",&llx_short,&lly_short,&urx_short,&ury_short,&n);
    if ((k != 4) || (llx_short != (short)123) || (lly_short != (short)456) ||
	(urx_short != (short)789) ||
	(ury_short != (short)987) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 1.6: tscanf() -> %d:%s",k,"");

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%d%n",&i,&n);
    if ((k != 1) || (i != 123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 2: tscanf() -> %d: %d",k,i);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%ld%n",&l,&n);
    if ((k != 1) || (l != 123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 3: tscanf() -> %d: %ld",k,l);

    f = (float)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%f%n",&f,&n);
    if ((k != 1) || (f != 123.0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 4: tscanf() -> %d: %f",k,f);

    d = (double)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%lf%n",&d,&n);
    if ((k != 1) || (d != 123.0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 5: tscanf() -> %d: %f",k,d);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "\t\t 123 \t\t"," %s%n",&s[0],&n);
    if ((k != 1) || (strcmp("123",s) != 0) || (n == UNSET) || (str[n] != ' '))
	ERROR("Test 6: tscanf() -> %d: %s",k,s);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%i%n",&i,&n);
    if ((k != 1) || (i != 123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 7: tscanf() -> %d: %d",k,i);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%li%n",&l,&n);
    if ((k != 1) || (l != 123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 8: tscanf() -> %d: %ld",k,l);

    n = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "ABCD 123","ABCD\t\t%n",&n);
    if ((k != 0) || (n != 5) || (n == UNSET) || (str[n] != '1'))
	ERROR("Test 9.1: tscanf() -> %d: %d",k,n);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "ABCD 123","ABCD\t\t%i%n",&i,&n);
    if ((k != 1) || (i != 123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 9.2: tscanf() -> %d: %d",k,i);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "ABCD123","ABCD\t\t%i%n",&i,&n);
    if ((k != 1) || (i != 123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 9.3: tscanf() -> %d: %d",k,i);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%o%n",&i,&n);
    if ((k != 1) || (i != 0123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 10: tscanf() -> %d: %d",k,i);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%lo%n",&l,&n);
    if ((k != 1) || (l != 0123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 11: tscanf() -> %d: %ld",k,l);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%x%n",&i,&n);
    if ((k != 1) || (i != 0x123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 10: tscanf() -> %d: %d",k,i);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "123","%lx%n",&l,&n);
    if ((k != 1) || (l != 0x123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 11: tscanf() -> %d: %ld",k,l);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0x123","%x%n",&i,&n);
    if ((k != 1) || (i != 0x123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 12: tscanf() -> %d: %d",k,i);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0x123","0x%x%n",&i,&n);
    if ((k != 1) || (i != 0x123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 12.1: tscanf() -> %d: %d",k,i);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0x123","%lx%n",&l,&n);
    if ((k != 1) || (l != 0x123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 13: tscanf() -> %d: %ld",k,l);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0x123","0x%lx%n",&l,&n);
    if ((k != 1) || (l != 0x123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 13.1: tscanf() -> %d: %ld",k,l);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0X123","%x%n",&i,&n);
    if ((k != 1) || (i != 0x123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 14: tscanf() -> %d: %d",k,i);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0X123","0%*[xX]%x%n",&i,&n);
    if ((k != 1) || (i != 0x123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 14.1: tscanf() -> %d: %d",k,i);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0X123","0X%x%n",&i,&n);
    if ((k != 1) || (i != 0x123) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 14.2: tscanf() -> %d: %d",k,i);

    i = (int)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0X123","%lx%n",&l,&n);
    if ((k != 1) || (l != 0x123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 15: tscanf() -> %d: %ld",k,l);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0X123","0%*[xX]%lx%n",&l,&n);
    if ((k != 1) || (l != 0x123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 15.1: tscanf() -> %d: %ld",k,l);

    l = (long)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "0X123","0X%lx%n",&l,&n);
    if ((k != 1) || (l != 0x123L) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 15.2: tscanf() -> %d: %ld",k,l);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "AAUUXX","%[AEIOU]%n",&s[0],&n);
    if ((k != 1) || (strcmp("AAUU",s) != 0) || (n == UNSET) || (str[n] != 'X'))
	ERROR("Test 16: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "XXUUAA","%[^AEIOU]%n",&s[0],&n);
    if ((k != 1) || (strcmp("XX",s) != 0) || (n == UNSET) || (str[n] != 'U'))
	ERROR("Test 17: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    s[4] = '\0';
    n = (int)UNSET;
    k = tscanf(str = " 123 ","%4c%n",&s[0],&n);
    if ((k != 1) || (strcmp(" 123",s) != 0) || (n == UNSET) || (str[n] != ' '))
	ERROR("Test 18: tscanf() -> %d: %s",k,s);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111","%ld%n",&pages[0],&n);
    check_pages(k,1,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222",
	       "%ld:%ld%n",
	       &pages[0],&pages[1],&n);
    check_pages(k,2,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222:333333",
	       "%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&n);
    check_pages(k,3,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222:333333:444444",
	       "%ld:%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&pages[3],&n);
    check_pages(k,4,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222:333333:444444:555555",
	       "%ld:%ld:%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&pages[3],&pages[4],&n);
    check_pages(k,5,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222:333333:444444:555555:666666",
	       "%ld:%ld:%ld:%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&pages[3],&pages[4],&pages[5],&n);
    check_pages(k,6,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222:333333:444444:555555:666666:777777",
	       "%ld:%ld:%ld:%ld:%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&pages[3],&pages[4],&pages[5],
	       &pages[6],&n);
    check_pages(k,7,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(str = "111111:222222:333333:444444:555555:666666:777777:888888",
	       "%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&pages[3],&pages[4],&pages[5],
	       &pages[6],&pages[7],&n);
    check_pages(k,8,pages,(int)str[n]);

    clear_pages(pages);
    n = (int)UNSET;
    k = tscanf(
    str = "111111:222222:333333:444444:555555:666666:777777:888888:999999",
	       "%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld%n",
	       &pages[0],&pages[1],&pages[2],&pages[3],&pages[4],&pages[5],
	       &pages[6],&pages[7],&pages[8],&n);
    check_pages(k,9,pages,(int)str[n]);

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    clear_string(s3,sizeof(s3));
    clear_string(s4,sizeof(s4));
    n = (int)UNSET;
    k = tscanf(str = "iamr10 1500 -> amr10 1499 % comment",
	       " %s %s -> %s %s%n",
	       s1, s2, s3, s4, &n);
    if (
	(k != 4) ||
	(strcmp("iamr10",s1) != 0) ||
	(strcmp("1500",  s2) != 0) ||
	(strcmp("amr10", s3) != 0) ||
	(strcmp("1499",  s4) != 0) ||
	(n == UNSET) || (str[n] != ' ')
	)
	ERROR("Test 20: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "a","%[a-c]%n",s,&n);
    if ((k != 1) || (strcmp("a",s) != 0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Standard C extension: Test 21.1: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "b","%[a-c]%n",s,&n);
    if ((k != 1) || (strcmp("b",s) != 0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Standard C extension: Test 21.2: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "c","%[a-c]%n",s,&n);
    if ((k != 1) || (strcmp("c",s) != 0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Standard C extension: Test 21.3: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    clear_string(s1,sizeof(s1));
    n = (int)UNSET;
    k = tscanf(str = "`","%[a-c]%n",s,&n);
    if ((k != 0) || (strcmp(s,s1) != 0))
	ERROR("Standard C extension: Test 21.4: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    clear_string(s1,sizeof(s1));
    n = (int)UNSET;
    k = tscanf(str = "d","%[a-c]%n",s,&n);
    if ((k != 0) || (strcmp(s,s1) != 0) || (n == UNSET) || (str[n] != 'd'))
	ERROR("Standard C extension: Test 21.5: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "hello, world","%[a-z]%n",s,&n);
    if ((k != 1) || (strcmp("hello",s) != 0) || (n == UNSET) || (str[n] != ','))
	ERROR("Standard C extension: Test 22: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "hello, world","%[a-z,]%n",s,&n);
    if ((k != 1) || (strcmp("hello,",s) != 0) || (n == UNSET) ||
	(str[n] != ' '))
	ERROR("Standard C extension: Test 23: tscanf() -> %d: %s",k,s);

    clear_string(s,sizeof(s));
    n = (int)UNSET;
    k = tscanf(str = "hello, world","%[a-z, ]%n",s,&n);
    if ((k != 1) || (strcmp("hello, world",s) != 0) || (n == UNSET) ||
	(str[n] != '\0'))
	ERROR("Standard C extension: Test 24: tscanf() -> %d: %s",k,s);

    f = (float)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "125","%t%n",&f,&n);
    if ((k != 1) || ((int)(1000.0 * f) != 125) || (n == UNSET) ||
	(str[n] != '\0'))
	ERROR("Test 25: tscanf() -> %d: %f",k,d);

    d = (double)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "125","%lt%n",&d,&n);
    if ((k != 1) || ((int)(1000.0 * d) != 125) || (n == UNSET) ||
	(str[n] != '\0'))
	ERROR("Test 26: tscanf() -> %d: %f",k,d);

    f = (float)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "125.","%t%n",&f,&n);
    if ((k != 1) || (f != 125.) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 27: tscanf() -> %d: %f",k,d);

    d = (double)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "125.","%lt%n",&d,&n);
    if ((k != 1) || (d != 125.) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 28: tscanf() -> %d: %f",k,d);

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    n = (int)UNSET;
    k = tscanf(str = "key=value","%S = %S%n",s1,s2,&n);
    if ((k != 2) || (strcmp(s1,"key") != 0) || (strcmp(s2,"value") != 0) ||
	(n == UNSET) || (str[n] != '\0'))
	ERROR("Test 29: tscanf() -> %d%s", k, "");

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    n = (int)UNSET;
    k = tscanf(str = "key = value","%S = %S%n",s1,s2,&n);
    if ((k != 2) || (strcmp(s1,"key") != 0) || (strcmp(s2,"value") != 0) ||
	(n == UNSET) || (str[n] != '\0'))
	ERROR("Test 30: tscanf() -> %d%s", k, "");

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    n = (int)UNSET;
    k = tscanf(str = "key = \"value with spaces\"","%S = %q%n",s1,s2,&n);
    if ((k != 2) || (strcmp(s1,"key") != 0) ||
	(strcmp(s2,"value with spaces") != 0) ||
	(n == UNSET) || (str[n] != '\0'))
	ERROR("Test 31: tscanf() -> %d%s", k, "");

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    n = (int)UNSET;
    k = tscanf(str = "key=\"value with spaces\"","%S = %q%n",s1,s2,&n);
    if ((k != 2) || (strcmp(s1,"key") != 0) ||
	(strcmp(s2,"value with spaces") != 0) ||
	(n == UNSET) || (str[n] != '\0'))
	ERROR("Test 32: tscanf() -> %d%s", k, "");

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    n = (int)UNSET;
    k = tscanf(str = "key = \"value with \"\"quotes\"\"\"","%S = %q%n",
	       s1,s2,&n);
    if ((k != 2) || (strcmp(s1,"key") != 0) ||
	(strcmp(s2,"value with \"quotes\"") != 0) ||
	(n == UNSET) || (str[n] != '\0'))
	ERROR("Test 33: tscanf() -> %d%s", k, "");

    clear_string(s1,sizeof(s1));
    clear_string(s2,sizeof(s2));
    n = (int)UNSET;
    k = tscanf(str = "key=\"value with \"\"quotes\"\"\"","%S = %q%n",s1,s2,&n);
    if ((k != 2) || (strcmp(s1,"key") != 0) ||
	(strcmp(s2,"value with \"quotes\"") != 0) ||
	(n == UNSET) || (str[n] != '\0'))
	ERROR("Test 34: tscanf() -> %d%s", k, "");

    for (k = 0; k < sizeof(fdimen)/sizeof(fdimen[0]); ++k)
	fdimen[k] = (float)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "100in 7200bp 7227pt 254cm 562.84592084cc 6754.1510501dd \
2540mm 602.25pc 473628672sp","%T %T %T %T %T %T %T %T %T%n",
	       &fdimen[0], &fdimen[1], &fdimen[2], &fdimen[3], &fdimen[4],
	       &fdimen[5], &fdimen[6], &fdimen[7], &fdimen[8], &n);
    if ((k != 9) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 35: tscanf() -> %d%s", k, "");
    else
    {
	for (k = 0; k < sizeof(fdimen)/sizeof(fdimen[0]); ++k)
	{
	    if ((int)(fdimen[k] + 0.5) != 7227)
		ERROR("Test 35: tscanf(): bad conversion in fdimen[%d] = %f",
		      k, fdimen[k]);
	}
    }

    for (k = 0; k < sizeof(ddimen)/sizeof(ddimen[0]); ++k)
	ddimen[k] = (double)UNSET;
    n = (int)UNSET;
    k = tscanf(str = "100in 7200bp 7227pt 254cm 562.84592084cc 6754.1510501dd \
2540mm 602.25pc 473628672sp","%lT %lT %lT %lT %lT %lT %lT %lT %lT%n",
	       &ddimen[0], &ddimen[1], &ddimen[2], &ddimen[3], &ddimen[4],
	       &ddimen[5], &ddimen[6], &ddimen[7], &ddimen[8], &n);
    if ((k != 9) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 36: tscanf() -> %d%s", k, "");
    else
    {
	for (k = 0; k < sizeof(ddimen)/sizeof(ddimen[0]); ++k)
	{
	    if ((int)(ddimen[k] + 0.5) != 7227)
		ERROR("Test 36: tscanf(): bad conversion in ddimen[%d] = %f",
		      k, ddimen[k]);
	}
    }

    for (k = 0; k < sizeof(idimen)/sizeof(idimen[0]); ++k)
	idimen[k] = UNSET;
    n = (int)UNSET;
    k = tscanf(str = "",
	       "%n %n %n %n %n %n %n %n %n %n",
	       &idimen[0], &idimen[1], &idimen[2], &idimen[3], &idimen[4],
	       &idimen[5], &idimen[6], &idimen[7], &idimen[8], &n);
    if ((k != 0) || (n == UNSET) || (str[n] != '\0'))
	ERROR("Test 37: tscanf() -> %d%s", k, "");
    else
    {
	for (k = 0; k < sizeof(idimen)/sizeof(idimen[0]); ++k)
	{
	    if (idimen[k] == UNSET)
		ERROR("Test 37: tscanf(): bad conversion in idimen[%d] = %f",
		      k, idimen[k]);
	}
    }

    exit(error_count ? EXIT_FAILURE : EXIT_SUCCESS);
    return(0);
}

static void
#if STDC
check_pages(
int	got,
int	expect,
long	pages[10],
int	next_char
)
#else /* NOT STDC */
check_pages(got,expect,pages,next_char)
int	got;
int	expect;
long	pages[10];
int	next_char;
#endif /* STDC */
{
    int k;

    for (k = 0; k < expect; ++k)
    {
	if ((got != expect) || (pages[k] != 111111L*(k+1)))
	    ERROR("Test 19: tscanf() -> %d: %ld",got,pages[k]);
    }
    if (next_char != '\0')
	ERROR("Test 19: tscanf() -> next_char = 0x%02x instead of NUL%s",
	      (unsigned int)next_char,"");
}

static void
#if STDC
clear_pages(
long	pages[10]
)
#else /* NOT STDC */
clear_pages(pages)
long	pages[10];
#endif /* STDC */
{
    int	k;

    for (k = 0; k < (int)sizeof(pages); ++k)
	pages[k] = (long)UNSET;
}

static void
#if STDC
clear_string(
char	*s,
size_t	n
)
#else /* NOT STDC */
clear_string(s,n)
char	*s;
size_t	n;
#endif /* STDC */
{
    (void)memset(s,0xff,n);
    s[n-1] = '\0';
}
#endif /* TEST */
